#!/usr/bin/perl -w
# 
# A script that gets a path to VM (as a parameter).
# Finds all VMDKs on the mounted with `VV mount` safeset 
# and mounts their guest file system(s) into the 
# specified destination: $gRootMountPoint 
# 
# The script supports MBR partions (Linux native and LVM)
# can mount ext2,ext3,ext4 filesystems.
# 
# Parameters:
#    1. Full path to virtual machine directory (required) 
#    2. Destination root mount directory (optional) 
# 
#
use strict;

use File::Path qw(make_path remove_tree);
use File::Basename;
use Getopt::Long;

my $gCacheDir = "/opt/cache";
my $gRootMountPoint = "/opt/share";

my $gRetOK = 0;
my $gRetError = 1;
my $gRetLvmConflict = 2;
my $gRetEmpty = 3;
my $gRetLdmDetected = 4;

# Turn autoflush on
BEGIN { $| = 1 }

############################################################
sub evlog
{
  print "log:", @_;
}

############################################################
sub everror
{
  print "err:", @_;
  exit $gRetError;
}

############################################################
sub evparam($$)
{
  my $paramname = shift;
  my $value = shift; 
  print "param:", $paramname, "=", $value, "\n";
}

############################################################
sub evuser
{
  print "userlog:", @_;
}

############################################################
# Escape any characters that could cause problems in bash 
# when within quotes
sub esc_bash_dbl_quote_chars {
  my $str = $_[0];
  $str =~ s/([`\$"])/\\$1/g;
  return $str;
}

############################################################
# We need to control output.
sub runcmd 
{
  my $retcode = -1;
  my $pid = open( my $fh, "-|", @_);
  if (defined($pid))
  {
    if ($pid > 0)
    {
      while ( defined (my $line = <$fh>)) 
      {
        chomp($line);
        evlog "$line\n";
      }

      close($fh);
      $retcode = $?;
    }
    # if pid == 0 it's a child, ignore it...
  }
  else
  {
    evlog "Can't execute command: $!\n";
  }
  return $retcode;
}

############################################################
sub has_ldm_partition($$)
{
  # has_ldm_partition: Tests whether any of the partitions in
  # a VMDK flat file are formatted as Windows Dynamic Disks. This
  # is done separately from get_partition_type() because the blkid
  # utility used there is incapable of detecting an LDM
  # partition.

  my $dir = shift;
  my $dfile = shift;
  open(my $vmdk, "<", "$dir/$dfile")
           or everror "Can't open $dfile for reading: $!\n";
  binmode($vmdk);

  my $MBR_SIZE = 512;
  my $PART_NUMBER = 4;
  my $PART_ENTRY_SIZE = 16;
  my $PART_TYPE_OFFSET = 4;
  my $PART_TABLE_OFFSET = 446;
  my $PART_WINDOWS_LDM_TYPE = 0x42;

  read($vmdk, my $mbr, $MBR_SIZE) == $MBR_SIZE
    or everror "Can't read MBR from $dfile: $!\n";

  close($vmdk);

  my $type_offset = $PART_TABLE_OFFSET + $PART_TYPE_OFFSET;

  for (my $i=1; $i<$PART_NUMBER + 1; $i++)
  {
    my $type = ord(substr($mbr, $type_offset, 1));
    if ($type == $PART_WINDOWS_LDM_TYPE)
    {
      evlog "Partition $i in $dfile is LDM.\n";
      return 1;
    }
    $type_offset += $PART_ENTRY_SIZE;
  }

  return 0;
}

############################################################
sub setup_loop_dev($$)
{
  my $dir = shift;
  my $dfile = shift;
  my $ldev;
  open(my $losetuppipe, "-|", "losetup",  ("-s",  "-f", "$dir/$dfile")) 
           or everror "Can't execute losetup for $dfile: $!\n";
  while ( defined ($ldev = <$losetuppipe>)) 
  {
    chomp($ldev);
    evlog "Added new loopback device: $ldev\n";
    last;
  }
  close($losetuppipe);
  return $ldev;
}


############################################################
sub create_cache_file($$)
{
  my $cachedir = shift;
  my $vmdkfile = shift;
  evlog "Creating cache for $vmdkfile ...\n";
  # Create cache dir 
  make_path($cachedir, {error => \my $err} );
  if (@$err) 
  {
    for my $diag (@$err) 
    {
      my ($file, $message) = %$diag;
      evlog "Error: $message\n";
    }
    everror "Failed to create mount point $cachedir.\n"; 
  }

  my $cache_file = $cachedir . "/" . $vmdkfile . ".cache";
  my $cache_file_escaped = esc_bash_dbl_quote_chars $cache_file;
  if (runcmd("dd if=/dev/zero of=\"$cache_file_escaped\" bs=128M count=1 2>&1") != 0)
  {
    everror "Can't create cache file $cache_file.\n";
  }
  return $cache_file;
}


############################################################
# Create device mapper snapshot
sub create_snapshot($$$)
{
  my $vmdk_dev = shift;
  my $cow_dev = shift;
  my $snapshot_name = shift;
  my $dsize = `blockdev --getsz $vmdk_dev`;
  chomp($dsize);
  my $syscmd = "echo 0 $dsize snapshot $vmdk_dev $cow_dev N 8 | dmsetup create $snapshot_name";
  if (runcmd($syscmd) != 0)
  {
    everror "Can't create dm snapshot $snapshot_name.\n";
  }

  return "/dev/mapper/" . $snapshot_name;
}


############################################################
sub map_partitions($)
{
  my @parts;
  my $ldev = shift;
  open( my $kpartxpipe, "-|", "kpartx",  ("-v",  "-a", "$ldev"))
           or everror "Can't execute kpartx for $ldev: $!\n";
  while ( defined (my $line = <$kpartxpipe>)) 
  {
    chomp($line);
    # Find loopback device in line that looks like:
    # 'add map loop8p1 : 0 208782 linear /dev/loop8 63'
    if ($line =~ /^add map ([0-9A-Fa-f]+\-disk[0-9]+p[0-9]+)/) 
    {
      my $part = "/dev/mapper/" . $1;
      evlog "Found new partition: '$part'\n";
      push(@parts, $part);
    }
  }
  close($kpartxpipe);
  return @parts;
}


############################################################
sub get_partition_type($)
{
  my $partition = shift;
  my $ptype = "unknown";
  my $blockinfo = `blkid $partition`;  
  if ($blockinfo =~ /\sTYPE=\"(.+)\"/)
  {
    $ptype = $1;
  }
  return $ptype;
}


############################################################
sub mount_volume($$$$$$)
{
  my $ptype = shift;
  my $device = shift;
  my $rootdir = shift;
  my $voldir = shift;
  my $islvm = shift;
  my $errinfo = shift; # Extra information for non-lvm unsupported error message
  
  # check if we can mount this type
  if ($ptype =~ /(ext2|ext3|ext4|ntfs|fat)/)
  {
    # Create mount point
    # Device will be converted from loop43p3 to disk43p3 
    # disk number may not correspond to the actual disk number
    my $mpoint = $rootdir . "/" . $voldir;
    evlog "Mounting $device type $ptype to $mpoint ...\n";
    make_path($mpoint, {error => \my $err} );
    if (@$err) 
    {
      for my $diag (@$err) 
      {
        my ($file, $message) = %$diag;
        evlog "Error: $message\n";
      }
      everror "Failed to create mount point $mpoint.\n";
    }
    # do mount
    runcmd("mount -t $ptype -r $device $mpoint 2>&1") == 0
      or everror "Mount of $device to $mpoint failed.\n";

    evlog "Done.\n";
    return 0;
  }
  else
  {
    # No error message for swap partitions
    if ($ptype !~ /^swap/)
    {
      # Error message for user log dependent on whether volume is LVM or not
      if ($islvm == 1)
      {
        evuser "unsupportedlvm-$device\n";
      }
      else
      {
        evuser "unsupported-$errinfo\n";
      }
      evlog "Skipping unsupported partition type $ptype on $device.\n";
    }
    return -1;
  }
}

############################################################
sub check_existing_lvm_conflict($)
{
  my $vmdir = shift;
  #check if there is any physical logical volume presenting in the system first
  my $pvscan_result = `pvscan 2>/dev/null`;
  if ($pvscan_result =~ /No matching physical volumes found/)
  {
    return;
  }
  opendir(my $dh, $vmdir) or everror "Can't open $vmdir: $!\n";
  my @files = readdir($dh);
  closedir($dh);
  foreach my $file (@files)
  {
    if ($file =~ /-flat\.vmdk$/ || $file =~ /-rdm\.vmdk$/)
    {
      my $line = `file -b \'$vmdir/$file\'`;
      chomp($line);
      my @partitions = split(/;/, $line);
      foreach my $partition (@partitions)
      {
        #find lvm partition by type ID=0x8e
        if ($partition =~ /^\s*partition\s+\d+:\s*(ID=0x8e),/)
        {
          evlog "Detected lvm partition in $file. Only single virtual machine with logical volume manager can be accessed at a time.\n";
          # Special case here to exit with code 2, to notify VV that another LVM is already mounted. 
          exit $gRetLvmConflict; 
        }
      }
    }
  }
}

############################################################
# MAIN

# Get the first argument
my $vmdir; 
# Check if mount point is specified
my $mountpoint;

GetOptions('vmdir=s' => \$vmdir,
           'mountpoint:s' => \$mountpoint);
unless (defined ($vmdir))
{
  evlog "USAGE: $0 --vmdir <path_to_vm_dir> [--mountpoint root_mount_point]\n";
  exit $gRetError;
}

# Init some variables
my $mountid; 
my $vmid = basename($vmdir);
if ($vmdir =~ /^\/opt\/mount\/([0-9a-fA-F]+)\//)
{
  $mountid = $1;
  evlog "Using mountid $mountid\n";
}
else
{
  everror "Cannot extract mount id.\n";
}

my $disknum = 1;
my $havelvm;
my $haveldm;
unless (defined($mountpoint))
{
  $mountpoint = $gRootMountPoint . "/" . $mountid . "/" . $vmid; 
}
#check if there is any conflict on lvm
check_existing_lvm_conflict($vmdir);

# Scan VM directory for vmdks
opendir(my $dh, $vmdir) or everror "Can't open $vmdir: $!\n";
my @files = readdir($dh);
closedir($dh);

my $num_volumes_mounted = 0;

foreach my $file (@files)
{
  # TODO: make it more robust, detecting file by name may not be a good idea  
  if($file =~ /-flat\.vmdk$/ || $file =~ /-rdm\.vmdk$/)
  {
    evlog "Inspecting file: " . $file . "\n";
    if (has_ldm_partition($vmdir, $file))
    {
      $haveldm = 1;
    }
    my $ldev = setup_loop_dev($vmdir, $file);
    my $cowfile = create_cache_file($gCacheDir."/".$mountid."/".$vmid, $file);
    if (defined($ldev) && defined($cowfile))
    {
      # Setup loopback device for cache file
      my $ldevcow = setup_loop_dev("", $cowfile);
      if (defined($ldevcow))
      {
        my $writable_vmdk_dev = create_snapshot($ldev, $ldevcow, 
                                                $mountid."-disk".$disknum);
        if (defined($writable_vmdk_dev))
        {
          my @parts = map_partitions($writable_vmdk_dev);
          my $partnum = 1;
          foreach my $part (@parts)
          {
            evlog "Processing partition $part\n";
            my $ptype = get_partition_type($part);
            if ($ptype =~ /^lvm/)
            {
              evlog "Partition $part appears to be LVM volume.\n";
              $havelvm = 1;
            }
            else
            {
              my $voldir = basename($part);
              $voldir =~ s/^[0-9A-Fa-f]+\-//;
              my $errinfo = $partnum . ',' . $file;
              if (mount_volume($ptype, $part, $mountpoint, $voldir, 0, $errinfo) == 0)
              {
                 $num_volumes_mounted++;
              }
            }
            $partnum++;
          } # foreach loop
        }
      }
    }
  }
  $disknum++;
}

# Process LVM volumes 
if (defined($havelvm))
{
  my $alllvms;
  runcmd("pvscan") == 0 or evlog "Failed to scan physical volumes.\n";
  runcmd("vgchange -ay 2>&1") == 0
    or evlog "Failed to activate volume groups.\n";
  open(my $lvscanpipe, "-|", "lvm lvscan 2>&1") 
    or evlog "Failed scan of logical volumes.\n";
  while(defined(my $line = <$lvscanpipe>))
  {
    chomp($line);
    if ($line =~ /^\s+ACTIVE\s+\'([\/\w\-]+)\'/)
    {
      my $lvmdev = $1;
      my $ptype = get_partition_type($lvmdev);
      my $voldir = $lvmdev;
      $voldir =~ s/^\/dev\///;
      $voldir =~ s/\//\-/;
      if (mount_volume($ptype, $lvmdev, $mountpoint, $voldir, 1, "") == 0)
      {
         $num_volumes_mounted++;
      }
      if (defined($alllvms))
      {
        $alllvms .= ",$voldir";
      }
      else
      {
        $alllvms = $voldir;
      }
    }
  }
  if (defined($alllvms))
  {
    evparam("--lvmdevices", $alllvms);
  }
  close($lvscanpipe);
}

# Exit with special return code to indicate that LDM partitions were
# detected
if ($haveldm)
{
  exit $gRetLdmDetected;
}

# Fail if no volumes were mounted
if ($num_volumes_mounted == 0)
{
  exit $gRetEmpty;
}
