#!/usr/bin/perl -w
#
# A script that gets a path to VM (as a parameter)
# from mounted with `VV mount` safeset.
# Finds all locally mounted guest file system(s) from this VM  
# based on path and unmounts them. Also releases all 
# resources a.e. loopback devices.
#
# Required parameters (see usage()):
#    1. VM mount directory with VMDKs
#
# Options
#    --force - try hard to unmount and release loopback devices 
#               even if they are used by some other process.
#

use strict;

use File::Path qw(make_path remove_tree);
use File::Basename;
use Getopt::Long;

my %args;
my %gLVMs;

# Turn autoflush on
BEGIN { $| = 1 }

sub term_handler {
    print("log: Received SIGTERM. Ignoring...\n");
}

$SIG{'TERM'} = 'term_handler';

############################################################
# Print log line.
sub evlog
{
  print "log:", @_;
}

############################################################
sub everror
{
  print "err:", @_;
  exit 1;
}

############################################################
# Print a parameter 
sub evparam($$)
{
  my $paramname = shift;
  my $value = shift; 
  print "param:", $paramname, " ", $value, "\n";
}

############################################################
# This is similar to system() but we control output 
# of the command.
sub runcmd 
{
  my $retcode = -1;
  my $pid = open( my $fh, "-|", @_);
  if (defined($pid))
  {
    if ($pid > 0)
    {
      while ( defined (my $line = <$fh>)) 
      {
        chomp($line);
        evlog "$line\n";
      }

      close($fh);
      $retcode = $?;
    }
    # if pid == 0 it's a child, ignore it...
  }
  else
  {
    evlog "Can't execute command: $!\n";
  }
  return $retcode;
}

############################################################
sub get_device_map($)
{
  my @mpdevs;
  my $umountid = shift;
  open( my $mountpipe, "-|", "mount") 
      or evlog "Can't execute mount: $!\n";
  while ( defined (my $line = <$mountpipe>)) 
  {
    chomp($line);
    # Find mount point from line that looks like:
    # /dev/mapper/61B5505c67A3-disk1p2 on .... type ext2 (rw)
    if ($line =~ /^\/dev\/mapper\/($umountid-disk[0-9]+)(p[0-9]+) on/)
    {
      my $diskdev = "/dev/mapper/" . $1;
      evlog "Disk dev: $diskdev\n";
      my $mpoint = $diskdev . $2;
      evlog "Mount point: $mpoint\n";
      push(@mpdevs, $mpoint);
    }
  }
  close($mountpipe);
  return @mpdevs;
}

############################################################
sub unmount($$)
{
  my $mpoint = shift;
  my $force = shift;  
  evlog "Unmounting: " . $mpoint . ".\n";
  if (runcmd("umount $mpoint 2>&1") != 0)
  {
    if ($force)
    {
      # if force flag is set, we need to kill all processes 
      # accessing our mount point and unmount device with 
      # lazy cleanup.
      runcmd("fuser", "-k", "-m", $mpoint); 
      runcmd("umount", "-l", $mpoint);
    }
    else
    {
      evlog "Cannot unmount $mpoint: $!\n";
    }
  }
}

############################################################
sub free_loopbacks($$)
{
  my $umountid = shift;
  my $vmid = shift;

  # Older losetup has a limit on a path to file...
  my $fileregex = "\/".$umountid."(\/.VSPHERE.)?\/". 
                   quotemeta(substr($vmid,0,24));
  open( my $losetupp, "-|", "losetup", "-a")
         or evlog "Can't execute losetup: $!\n";
  while ( defined (my $line = <$losetupp>))
  {
    chomp($line);
    # Find loopback device from line that looks like:
    # /dev/loop1: [001e]:4 (path to a file)...
    if (($line =~ /$fileregex/) && ($line =~ /^(\/dev\/loop[0-9]+)/))
    {
      my $device = $1;
      evlog "Freeing loopback: $device\n";
      runcmd("losetup -d $device 2>&1") == 0 
        or evlog "Can't remove device $device\n";
    }
  }
  close($losetupp);
}

############################################################
sub unmap_kpartx_devices($)
{
  my $umountid = shift;
  if (opendir(my $dmh, "/dev/mapper"))
  {
    my @files = readdir($dmh);
    closedir($dmh);
    foreach my $file (@files)
    {
      if($file =~ /^$umountid-disk[0-9]+$/)
      {
        evlog "Unmapping device: '$file'\n";
        runcmd("kpartx -d /dev/mapper/$file 2>&1") == 0
          or evlog "Can't execute kpartx for $file\n";
      }
    }
  }
  else
  {
    evlog "Can't open /dev/mapper directory: $!\n";
  }
}

############################################################
# Unmap partition devices that mach pattern: 
#                  /dev/mapper/<mountid>-diskNpN
sub unmap_partition_devices($)
{
  my $umountid = shift;
  if (opendir(my $dmh, "/dev/mapper"))
  {
    my @files = readdir($dmh);
    closedir($dmh);
    foreach my $file (@files)
    {
      if($file =~ /^$umountid-disk[0-9]+p[0-9]+$/)
      {
        evlog "Found dm node to be removed:" . $file . "\n";
        runcmd("dmsetup remove -f /dev/mapper/$file 2>&1") == 0
            or evlog "Failed to unmap node $file: $!\n";
      }
    }
  }
  else
  {
    evlog "Can't open /dev/mapper directory: $!\n";
  }
}

############################################################
# Unmapi disk devices that mach pattern: 
#                  /dev/mapper/<mountid>-diskNpN
sub unmap_disk_devices($)
{
  my $umountid = shift;
  if (opendir(my $dmh, "/dev/mapper"))
  {
    my @files = readdir($dmh);
    closedir($dmh);
    foreach my $file (@files)
    {
      if($file =~ /^$umountid-disk[0-9]+$/)
      {
        evlog "Found dm node to be removed:" . $file . "\n";
        runcmd("dmsetup remove -f /dev/mapper/$file 2>&1") == 0
            or evlog "Failed to unmap node $file: $!\n";
      }
    }
  }
  else
  {
    evlog "Can't open /dev/mapper directory: $!\n";
  }
}

############################################################
sub unmount_directories($$)
{
  my $mountpoint = shift;
  my $force = shift;
  my %gLVMs = ();
  # For safety sake, check the mount point before we start 
  # unmouning
  unless ($mountpoint =~ /^\/opt\/share\//)
  {
    evlog "Invalid mount point $mountpoint!\n";
    return;
  }
  if (opendir(my $d, $mountpoint))
  {
    my @files = readdir($d);
    closedir($d);
    foreach my $file (@files)
    {
      next if($file =~ /^\.$/);
      next if($file =~ /^\.\.$/);

      unmount($mountpoint . "/" . $file, $force);  
      # Save LVM device for later cleanup.
      unless($file =~ /^disk[0-9]+p[0-9]+$/)
      {
        $gLVMs{$file} = 1;
      }
    }
  }
  else
  {
    evlog "Can't open $mountpoint: $!\n";
  }
}

############################################################
# Remove a directory with all subdirs 
sub evremove_tree($)
{
  my $rdir = shift;
  if ( -d $rdir)
  {
    evlog "Removing directory: $rdir\n";
    remove_tree( $rdir, {error => \my $err} );
    if (@$err) 
    {
      for my $diag (@$err) 
      {
        my ($file, $message) = %$diag;
        evlog "Error: $message\n";
      }
    }
  }
}

############################################################
# MAIN

# Flag, if set forces unmount
my $force = 0;
# Directory with VMDKs
my $vmdir;
# Root directory that contains all mount points to be removed
my $rootmountpoint;
# Coma separated list of LVM devce names
my $alllvms;
# Parse command line
GetOptions('force' => \$force,
           'vmdir=s' => \$vmdir,
           'lvmdevices:s' => \$alllvms,
           'mountpoint:s' => \$rootmountpoint);
 
unless (defined ($vmdir))
{
  evlog "USAGE: $0 [--force] [--lvmdevices lvmdevs] --vmdir vmdir [--mountpoint mountdirectory]\n";
  exit 2;
}

# Init some variables
my $mountid;
my $vmid = basename($vmdir);
if ($vmdir =~ /^\/opt\/mount\/([0-9a-fA-F]+)\//)
{
  $mountid = $1;
  evlog "Using mountid $mountid\n";
}
else
{
  everror "Cannot extract mount id.\n";
}
unless (defined($rootmountpoint))
{
  $rootmountpoint = "/opt/share/$mountid/$vmid";
}

if (defined $alllvms)
{
  my @lvms = split(',', $alllvms);
  foreach my $val (@lvms) 
  {
    # Note we also discover lvm volumes from 
    # mounted directories in unmount_directories();
    $gLVMs{$val} = 1;
  }
}

runcmd("sync");
# 1. Unmount all mounted filesystem under VM directory
#-----------------------------------------------------------
unmount_directories($rootmountpoint, $force);
my @mpoints = get_device_map($mountid);
foreach my $mpoint (@mpoints)
{
  unmount($mpoint, $force);
}

# 2. Free LVM devices
#-----------------------------------------------------------
runcmd("vgchange -an 2>/dev/null");
foreach my $lvmd (keys %gLVMs)
{
  my $lvmname = "/dev/mapper/$lvmd";
  evlog "Removing lvm node: $lvmname.\n";
  if ( -e $lvmname )
  {
    runcmd("dmsetup", "remove", "-f", $lvmname); 
  }
}

# 3. Free device maps 
#-----------------------------------------------------------
unmap_kpartx_devices($mountid);
# Just make sure that all our mapped devices are gone.
unmap_partition_devices($mountid);
unmap_disk_devices($mountid);

# 4. Free loopback devices 
#-----------------------------------------------------------
free_loopbacks($mountid, $vmid);

# 5. Remove directories that were used to mount volumes 
#-----------------------------------------------------------
evremove_tree($rootmountpoint);
my $cachedir = "/opt/cache/$mountid";
evremove_tree($cachedir);
my $sharedir = "/opt/share/$mountid";
evremove_tree($sharedir);

